﻿# -*- coding: utf-8 -*-

import os, sys, datetime, shutil, tempfile
reload(sys)
sys.setdefaultencoding(sys.getfilesystemencoding())
sys.path.append(r'C:\PCI Geomatics\Geomatica 2016\exe')
import pci
from pci.api import datasource
from pci.iii import iii
from pci.link import link
from pci.pcimod import pcimod
from pci.fexport import fexport
from pci.pansharp2 import pansharp2
from pci.nspio import Counter, enableDefaultCounter, enableDefaultReport

# 센서별 Pixel Shifting Values 설정부 (X, Y)
K3_shift = (2,2)
K3A_shift = (0,0)

def now():
    '''Return the current time in this format: 2016-03-02 10:59:14.182'''
    #return time.strftime("%Y-%m-%d %H:%M:%S")
    return ('%s' % datetime.datetime.now())[:-7]

def getTempName(location=None, extension=''):
    if extension <> '':
        extension = '.' + extension
    filename = tempfile.NamedTemporaryFile(suffix=extension, dir=location).name
    return filename

def findPANscenefile(targetpath):
    files = os.listdir(targetpath)
    PANscenefile = ''
    for file in files:
        if file.upper()[-8:] == '_AUX.XML' and getSensor(file) in ['K3','K3A']:
            PANscenefile = os.path.join(targetpath, file + '-PAN')
            break
        elif file.upper()[-4:] == '.EPH' and getSensor(file) == 'MSC' and file.upper()[31] == 'P':
            PANscenefile = os.path.join(targetpath, file)
            break
    return PANscenefile

def getSensor(scenefilename):
    '''Get sensor name from KOMPSAT scene file name: MSC/K3/K3A'''
    scenefilename = os.path.basename(scenefilename)
    sensor = scenefilename.split('_')[0].upper()
    if sensor not in ['MSC','K3','K3A']:
        sensor = ''
    return sensor

def getLevel(scenefilename):
    '''Get level from KOMPSAT scene file name: 1R/1G'''
    sensor = getSensor(scenefilename)
    level = ''
    if sensor in ['K3','K3A']:
        pos = scenefilename.upper().find('_AUX.XML')
        level = scenefilename[pos-2:pos]
    elif sensor == 'MSC':
        pos = scenefilename.upper().find('.EPH')
        level = scenefilename[pos-2:pos]
    return level

def getMSCTilt(scenefilename):
    '''Get tilt value from KOMPSAT-2 (MSC) scene file name: Pxx/Nxx'''
    scenefilename = os.path.basename(scenefilename)
    tilt = ''
    if getSensor(scenefilename) == 'MSC':
        if scenefilename[31] == 'P':
            tilt = scenefilename[32:35]
        else:
            tilt = scenefilename[33:36]
    return tilt

def copyGeoinfo(inputfile, outputfile):
    '''Copy geoinfo from input to output'''
    indataset = datasource.open_dataset(inputfile)
    reader = datasource.BasicReader(indataset)
    outdataset = datasource.open_dataset(outputfile, datasource.AccessMode.eAM_WRITE)
    writer = datasource.BasicWriter(outdataset)
    writer.geocoding = reader.geocoding
    del writer, reader, outdataset, indataset
    
def getXYC(imagefilename):
    '''Get XSize(Columns), YSize(Rows), Channels of image'''
    try:
        dataset = datasource.open_dataset(imagefilename)
        return (dataset.width, dataset.height, dataset.chan_count)
    except:
        return None

def getauxfiles(scenefilename):
    '''Get auxfile list from PAN scene name'''
    sensor = getSensor(scenefilename)
    scenefilename = os.path.basename(scenefilename)
    if sensor in ['K3','K3A']:
        sceneid = scenefilename[:-12]+ '_P_rpc.txt'
        sceneid + '_P_rpc.txt'
    elif sensor == 'MSC':
        sceneid = scenefilename[:31]
    return sceneid

def getRawTIF(scenefilename):
    '''Get raw TIF file name from PAN scene name'''
    sensor = getSensor(scenefilename)
    if sensor in ['K3','K3A']:
        if scenefilename[-3:] == 'PAN':
            scenefilename = scenefilename[:-12] + '_P.tif'
        elif scenefilename[-2:] == 'MS':
            scenefilename = scenefilename[:-11] + '_B.tif'
    elif sensor == 'MSC':
        scenefilename = scenefilename[:-4] + '.tif'
    return scenefilename
    
def getMSscene(scenefilename):
    '''Get MS scene name from PAN scene name'''
    sensor = getSensor(scenefilename)
    if sensor in ['K3','K3A']:
        scenefilename = scenefilename[:-3] + 'MS'
    elif sensor == 'MSC':
        scenepath, scenefilename = os.path.split(scenefilename)
        scenefilename = scenefilename[:31] + 'M1' + getMSCTilt(scenefilename) + 'G_' + getLevel(scenefilename) + '.eph'
        scenefilename = os.path.join(scenepath, scenefilename)
    return scenefilename

def getPSHscene(scenefilename):
    '''Get PSH scene name from PAN scene name'''
    sensor = getSensor(scenefilename)
    scenefilename = os.path.basename(scenefilename)
    if sensor in ['K3','K3A']:
        scenefilename = scenefilename[:-12]
    elif sensor == 'MSC':
        scenefilename = scenefilename[:31]
    return scenefilename + '_PS.pix'
    
def CreateDir(targetpath):
    '''Create Dir'''
    if not os.path.exists(targetpath):
        os.makedirs(targetpath)

def pansharpening(indir, outdir):
    indir = unicode(indir)
    outdir = unicode(outdir)
    
    enableDefaultCounter(mode='OFF')
    # Step 0 : Checking and preparing
    print 'Source Directory: ', indir
    print 'Output Directory: ', outdir
    print 'Checking input and output directory...'
    if not os.path.isdir(indir):
        print 'Error for checking input directory'
        return 1
    else:
        panfile = findPANscenefile(indir)
        if panfile == '':
            print 'No KOMPSAT scene found'
            return 1
        else:
            try:
                CreateDir(outdir)
            except:
                print 'Error for checking output directory'
                return 1
        try:
            sensor = getSensor(panfile)
            msfile = getMSscene(panfile)
            mstmpfile = getTempName(outdir, 'pix')
            link(msfile, mstmpfile, [])
            pantmpfile = getTempName(outdir, 'pix')
            link(panfile, pantmpfile, [])
            pshtmpfile = os.path.join(outdir, getPSHscene(panfile))

            if getLevel(panfile) == '1G':
                copyGeoinfo(getRawTIF(panfile), pantmpfile)
                copyGeoinfo(getRawTIF(msfile), mstmpfile)
                pass
            print '%s scene was found' % sensor
            print
        except:
            print 'Error for reading KOMPSAT scenes'
            try:
                os.remove(mstmpfile)
                os.remove(pantmpfile)
            except:
                pass
            return 1
    
    enableDefaultCounter(mode='ON')
    enableDefaultReport(output='OFF')
    # Step 1 : Adjusting PAN pixels
    if sensor in ['K3','K3A']:
        print '[1/6] Adjusting pixels in PAN temp file...'
        if sensor == 'K3':
            shiftx, shifty = K3_shift
        elif sensor == 'K3A':
            shiftx, shifty = K3A_shift
        pcimod(pantmpfile, 'ADD', [0,0,1])
        cols, rows, chans = getXYC(pantmpfile)
        cols, rows = cols-shiftx, rows-shifty
        iii(fili=pantmpfile, filo=pantmpfile, dbic=[1], dboc=[2], dbiw=[0,0,cols,rows], dbow=[shiftx,shifty,cols,rows])
        dbic = [1,2,3,4]
    else:
        dbic = [2,1,4,3]

    # Step 2 : Pansharpening
    print '[%d/%d] Pansharpening...' % ((1,4) if sensor == 'MSC' else (2,6))
    fili = mstmpfile
    dbic_ref = [1,2,3,4]
    fili_pan = pantmpfile
    cols, rows, chans = getXYC(pantmpfile)
    dbic_pan = [chans]
    filo = pshtmpfile
    dboc = []
    enhance = 'NO'
    srcbgd = 'FILE, 0'
    poption = 'OFF'
    ftype = 'PIX'
    foptions = ''
    pansharp2(fili, dbic, dbic_ref, fili_pan, dbic_pan, filo, dboc, enhance, srcbgd, poption, ftype, foptions)

    # Step 3 : Adjusting Pansharp pixels
    if sensor in ['K3','K3A']:
        print
        print '[3/6] Adjusting pixels in Pansharpened temp file...'
        cols, rows, chans = getXYC(pshtmpfile)
        cols, rows = cols-shiftx, rows-shifty
        pcimod(pshtmpfile, 'ADD', [0,0,chans])
        iii(fili=pshtmpfile, filo=pshtmpfile, dbic=[1,2,3,4], dboc = [5,6,7,8], dbiw=[shiftx, shifty, cols, rows], dbow=[0,0,cols,rows])
        pcimod(pshtmpfile, 'DEL', [1,2,3,4])

    # Step 4 : Convert Pansharp to TIF
    print '[%d/%d] Converting output to TIF files...' % ((2,4) if sensor == 'MSC' else (4,6))
    psoutfile = pshtmpfile[:-3] + 'tif'

    try:
        fexport(fili=pshtmpfile, filo=psoutfile, dbiw=[], dbic=[1,2,3,4], dbib=[], dbvs=[], dblut=[], dbpct=[], ftype='TIF', foptions='BigTIFF=NEVER')
    except:
        for chan in [1,2,3,4]:
            psoutfile = pshtmpfile[:-5] + 'BGRN'[chan-1] + '.tif'
            fexport(fili=pshtmpfile, filo=psoutfile, dbiw=[], dbic=[chan], dbib=[], dbvs=[], dblut=[], dbpct=[], ftype='TIF', foptions='BigTIFF=NEVER')

    # Step 5 : Copy auxiliary files
    print '[%d/%d] Copy auxiliary files...' % ((3,4) if sensor == 'MSC' else (5,6))
    for file in os.listdir(indir):
        temp = file.upper()
        if temp[-4:] in ['.JPG','.JGW']:
            shutil.copy(os.path.join(indir, file), outdir)
        elif sensor in ['K3','K3A'] and temp in [os.path.basename(panfile)[:-4].upper(), os.path.basename(panfile)[:-12].upper()+'_P_RPC.TXT']:
            shutil.copy(os.path.join(indir, file), outdir)
        elif sensor == 'MSC' and temp in [os.path.basename(panfile)[:-3].upper()+'EPH', os.path.basename(panfile)[:-3].upper()+'RPC', os.path.basename(panfile)[:-3].upper()+'TXT']:
            shutil.copy(os.path.join(indir, file), outdir)
    
    # Step 6 : Delete unnecessary files
    print '[%d/%d] Deleting unnecessary files...' % ((4,4) if sensor == 'MSC' else (6,6))
    for file in os.listdir(outdir):
        temp = file.upper()
        if temp[-4:] == '.POX':
            os.remove(os.path.join(outdir, file))
    os.remove(pantmpfile)
    os.remove(mstmpfile)
    os.remove(pshtmpfile)
    
    print
    print 'Completed.'
    return 0
    
def monitoring():
    work_dir = os.path.abspath('WORK')
    waiting_dir = os.path.abspath(work_dir + '/WAITING')
    running_dir = os.path.abspath(work_dir + '/RUNNING')
    completed_dir = os.path.abspath(work_dir + '/COMPLETED')
    failed_dir = os.path.abspath(work_dir + '/FAILED')
    CreateDir(waiting_dir)
    CreateDir(running_dir)
    CreateDir(failed_dir)
    CreateDir(completed_dir)

    print '[%s] %s Ready...' % (now(), 'KOMPSAT Pan Sharpening')
    
    while True:
        errorflag = 0
        # Check RUNNING dir
        running_jobs = os.listdir(running_dir)
        while len(running_jobs)>0:
            running_job = running_jobs.pop(0)
            print '[%s] Job Starting: %s' % (now(), running_job)
            errorflag = runProcess(os.path.join(running_dir, running_job))
            tmpjobname = running_job
            if errorflag == 1:
                ## 파일명 중복체크
                while os.path.isfile(failed_dir + '\\' + tmpjobname):
                    tmpjobname += "'"
                shutil.move(os.path.join(running_dir, running_job), os.path.join(failed_dir, tmpjobname))
                print
                print '[%s] Job Failed' % now()
                print
            else:
                ## 파일명 중복체크
                while os.path.isfile(completed_dir + '\\' + tmpjobname):
                    tmpjobname += "'"
                shutil.move(os.path.join(running_dir, running_job), os.path.join(completed_dir, tmpjobname))
                print
                print '[%s] Job Completed' % now()
                print
            print '[%s] %s Ready...' % (now(), 'KOMPSAT Pan Sharpening')
        
        # Check WAITING dir
        waiting_jobs = os.listdir(waiting_dir)
        if len(waiting_jobs)>0:
            shutil.move(os.path.join(waiting_dir, waiting_jobs.pop(0)), running_dir)
        
def runProcess(jobfile):
    params = dict()
    try:
        for i in open(jobfile):
            param, value = i.rstrip().split(': ')
            params[param] = value
        errorflag = pansharpening(params['Source'], params['Output'])
    except:
        errorflag = 1
    return errorflag
    
monitoring()